#ifndef _MDARRAYS_H
#define _MDARRAYS_H

#include<dan_math/types.h>
#include<stdio.h>
#include<stdbool.h>
#include<stdarg.h>

// Define a new gpointer type here to use in mdarrays
// because we don't know if gpointer will be typedef'ed by
// here or not.
typedef void * gpointer_l;

#define MDARRAYS_MAX_STRING_LENGTH  120
#define MDARRAYS_ARRAY_NAME_SIZE    30
#define MDARRAYS_MAX_DIMENSIONS     10

// The printf format types
#define MDARRAYS_PRINTF_TYPE_c      0
#define MDARRAYS_PRINTF_TYPE_d      1
#define MDARRAYS_PRINTF_TYPE_ld     2
#define MDARRAYS_PRINTF_TYPE_e      3
#define MDARRAYS_PRINTF_TYPE_Le     4
#define MDARRAYS_PRINTF_TYPE_E      5
#define MDARRAYS_PRINTF_TYPE_LE     6
#define MDARRAYS_PRINTF_TYPE_f      7
#define MDARRAYS_PRINTF_TYPE_Lf     8
#define MDARRAYS_PRINTF_TYPE_g      9
#define MDARRAYS_PRINTF_TYPE_Lg     10
#define MDARRAYS_PRINTF_TYPE_G      11
#define MDARRAYS_PRINTF_TYPE_LG     12
#define MDARRAYS_PRINTF_TYPE_p      13
#define MDARRAYS_PRINTF_TYPE_x      14
#define MDARRAYS_PRINTF_TYPE_X      15
#define MDARRAYS_PRINTF_TYPE_lX     16
#define MDARRAYS_PRINTF_TYPE_u      17

// These _access_type variables are used to set the type of arrays.
// They are global constants because the bounds checking preprocessor
// sees the code after the C preprocessor, making preprocessor defines
// impossible.  The values are set in access.c
extern const uint _mdarrays_access_type_real;
extern const uint _mdarrays_access_type_long_double;
extern const uint _mdarrays_access_type_double;
extern const uint _mdarrays_access_type_float;
extern const uint _mdarrays_access_type_long_int;
extern const uint _mdarrays_access_type_int;
extern const uint _mdarrays_access_type_short_int;
extern const uint _mdarrays_access_type_unsigned_long_int;
extern const uint _mdarrays_access_type_unsigned_int;
extern const uint _mdarrays_access_type_unsigned_short_int;
extern const uint _mdarrays_access_type_char;
extern const uint _mdarrays_access_type_unsigned_char;

// For structures or other objects, it is impossible to know what they will be 
// called ahead of time, and
// thus impossible to put an access type above.  Instead, use
// this generic "pointer" type, which will allow you to create an array of pointers.
extern const uint _mdarrays_access_type_pointer;

// When using variable arguments lists (as with _mdarrays_access()), the last argument in the
// list must indicate that the end of the arguments list has been reached
// I don't use -1 because I would like to be able to catch off-by-one errors when zero should be the
// lowest index, and -1 is used instead.  This marker is made a global variable instead of a compiler
// define because it is added by the bounds checking preprocessor after the normal cpp
// preprocessing step, and thus can't be set by the cpp preprocessor.  The value is set in access.c
// The variable should only be used by bounds_checking.pl
extern const uint ARRAY_LAST_INDEX_MARKER;
 
// These are the "type"s for real-array load and save routines.
// If this list of basic types is changed, update the docs.
// "bool" basic type is also allowed, it is just int internally.
#define MDARRAYS_CHAR_TYPE          0
#define MDARRAYS_UCHAR_TYPE         1
#define MDARRAYS_BOOL_TYPE          2
#define MDARRAYS_INT_TYPE           3
#define MDARRAYS_UINT_TYPE          4
#define MDARRAYS_LONG_INT_TYPE      5
#define MDARRAYS_FLOAT_TYPE         6
#define MDARRAYS_DOUBLE_TYPE        7
#define MDARRAYS_LONG_DOUBLE_TYPE   8
#define MDARRAYS_REAL_TYPE          9
#define MDARRAYS_GPOINTER_TYPE      10

// A NONMDARRAY_ARRAY is an array defined like:
// int a[2][3];
// A MDARRAY_ARRAY is an array defined like:
// int **a;
// These are the same thing for one dimensional array, but for higher
// dimensional arrays, they are different.
//

// This is put in the array info structure to make sure that an array being passed
// to array processing routines is really a mdarray
#define ARRAY_INFO_STRUCT_TAG  1234567890L

struct print_style {
    uint type;            // the T in %W.PT inside a print string
    uint width;           // the W in %W.PT inside a print string
    uint precision;       // the P in %W.PT inside a print string
    char delimiter;       // A character to print between elements when printing
    char line_delimiter;  // A character to print between elements when printing
};
typedef struct print_style print_style_struct;

// An mdarray_info_struct holds information in the memory preceding an array
struct mdarray_info {
    uint tag;               // Set to ARRAY_INFO_STRUCT_TAG for checking
    uint dimensions;        // Number of dimensions in the array
    uint type;              // One of the MDARRAYS_*_TYPE from this file
    uint type_size;         // Number of bytes for each element
    uint sizes[MDARRAYS_MAX_DIMENSIONS]; // Number of elements in each dimension
    uint elements;          // Total number of elements
    char name[MDARRAYS_ARRAY_NAME_SIZE]; // String name of the array
    print_style_struct print_style;
    print_style_struct save_style;
};
typedef struct mdarray_info mdarray_info_struct;

/***********************************/
//   Array allocation and deallocation routines
/***********************************/
// Note: these arrays are type POINTER_ARRAY for use with other routines
//       in this library, because they are pointers to pointers, instead
//       of a true grid array.
void * _mdarrays_allocate_macro_fill(char *cfile, uint cline, char *type_string, 
                                     uint type_size, void *value_pointer, bool is_object, ...);
void * _mdarrays_allocate_macro_nofill(char *cfile, uint cline, 
                                       char *type_string, uint type_size, bool is_object, ...);
void * _mdarrays_allocate_macro_fill_zero(char *cfile, uint cline, char *type_string, 
                                          uint type_size, bool is_object, ...);
void _mdarrays_deallocate(char *cfile, uint cline, void * array);
#define MDARRAYS_ALLOCATE(type, ...) \
   _mdarrays_allocate_macro_nofill(__FILE__, __LINE__, #type, sizeof(type), \
                                   false/*is_object*/, __VA_ARGS__, 0)
#define MDARRAYS_ALLOCATE_AND_CLEAR(type, ...) \
   _mdarrays_allocate_macro_fill_zero(__FILE__, __LINE__, #type, sizeof(type), \
                                      false/*is_object*/, __VA_ARGS__, 0)
#define MDARRAYS_ALLOCATE_AND_FILL(type, value_pointer, ...) \
   _mdarrays_allocate_macro_fill(__FILE__, __LINE__, #type, sizeof(type), \
                                 (void *)value_pointer, false/*is_object*/, __VA_ARGS__, 0)
#define MDARRAYS_ALLOCATE_OBJECT(type, ...) \
   _mdarrays_allocate_macro_nofill(__FILE__, __LINE__, #type, sizeof(type), \
                                   true/*is_object*/, __VA_ARGS__, 0)
#define MDARRAYS_ALLOCATE_AND_CLEAR_OBJECT(type, ...) \
   _mdarrays_allocate_macro_fill_zero(__FILE__, __LINE__, #type, sizeof(type), \
                                      true/*is_object*/, __VA_ARGS__, 0)
#define MDARRAYS_ALLOCATE_AND_FILL_OBJECT(type, value_pointer, ...) \
   _mdarrays_allocate_macro_fill(__FILE__, __LINE__, #type, sizeof(type), \
                                 (void *)value_pointer, true/*is_object*/, __VA_ARGS__, 0)

#define MDARRAYS_FREE(array) _mdarrays_deallocate(__FILE__, __LINE__, (void *)array)

// =================================
// Access routines
// =================================
typedef void (*mdarrays_foreach_function)(void *array, uint *indices, void *user_data);
#define MDARRAYS_FOREACH(array, foreach_function, user_data) \
    _mdarrays_foreach(__FILE__, __LINE__, array, foreach_function, user_data)
void _mdarrays_foreach(char *cfile, uint cline, void *array, 
                       mdarrays_foreach_function, void *user_data);
void _assign_pointer(char *cfile, uint cline, 
                     void *array, mdarray_info_struct *mdarray_info, char *array_name,
                     uint access_index[MDARRAYS_MAX_DIMENSIONS], void *value_pointer);
void _mdarrays_assign(char *cfile, uint cline, void *array, char *array_name, void *value_pointer, ...);
#define MDARRAYS_ASSIGN(array, value_pointer, ...) \
    _mdarrays_assign(__FILE__, __LINE__, array, #array, (void *)value_pointer, __VA_ARGS__)

// =================================
// Bounds checking and array info routines
// =================================
mdarray_info_struct *
_mdarrays_get_mdarray_info(char *cfile, uint cline, void *array, const char *array_name);
bool _mdarrays_is_mdarray(char *cfile, uint cline, void *void_pointer);
void * _mdarrays_access(char *cfile, uint cline, void *array, char *array_name, ...);
void _mdarrays_assert_bounds(char *cfile, uint cline, void *array, const char *array_name, ...);
#define MDARRAYS_IS_MDARRAY(array)  _mdarrays_is_mdarray(__FILE__, __LINE__, array)
#define MDARRAYS_ASSERT_BOUNDS(array, ...)  \
    _mdarrays_assert_bounds(__FILE__, __LINE__, array, #array, __VA_ARGS__)


/***********************************/
//   Array load and save routines
/***********************************/
#define MDARRAYS_SAVE(filename, description, array) \
        _mdarrays_save(__FILE__, __LINE__, filename, description, array)
void _mdarrays_save(char *cfile, uint cline, char *filename, char *description, void * array);
#define MDARRAYS_SAVE_TAGGED(filename, description, array, tags) \
        _mdarrays_save_tagged(__FILE__, __LINE__, filename, description, array, tags)
void _mdarrays_save_tagged(char *cfile, uint cline, char *filename, char *description,
                           void * array, char **tags);
#define MDARRAYS_LOAD(filename, description, array) \
        _mdarrays_load(__FILE__, __LINE__, filename, description, array)
void _mdarrays_load(char *cfile, uint cline, char *filename, char **description, void * array);
#define MDARRAYS_PRINT(file, array) \
        _mdarrays_print(__FILE__, __LINE__, file, array)
void _mdarrays_print(char *cfile, uint cline, FILE *file, void *array);
#define MDARRAYS_PRINT_TAGGED(file, array, tags) \
        _mdarrays_print_tagged(__FILE__, __LINE__, file, array, tags)
void _mdarrays_print_tagged(char *cfile, uint cline, 
                            FILE *file, void *array, char **tags);
// type and typesize are both needed because it's impossible to tell automatically
// what the size of a gpointer is.
#define MDARRAYS_SAVE_NONMDARRAY(filename, array, type, type_size, dimensions, sizes, \
                                 delimiter, line_delimiter, print_type, print_width, \
                                 print_precision) \
        _mdarrays_save_nonmdarray(__FILE__, __LINE__, \
                                  filename, array, type, type_size, dimensions, sizes, \
                                  delimiter, line_delimiter, print_type, print_width, \
                                  print_precision)
void _mdarrays_save_nonmdarray(char *cfile, uint cline, 
                               char *filename, void * array, uint type, unsigned
                               int type_size, uint dimensions, uint *sizes,
                               char delimiter, char line_delimiter,
                               uint print_type, uint print_width, uint print_precision);
#define MDARRAYS_LOAD_NONMDARRAY(filename, array, type, type_size, dimensions, sizes) \
        _mdarrays_load_nonmdarray(__FILE__, __LINE__, filename, array, type, type_size, dimensions, sizes) 
void _mdarrays_load_nonmdarray(char *cfile, uint cline, 
                               char *filename, void * array, uint type, 
                               uint type_size, uint dimensions, uint *sizes);
#define MDARRAYS_LOAD_2D_DYNAMIC(filename, type, rows, columns, description, status) \
        _mdarrays_load_2d_dynamic(__FILE__, __LINE__, filename, #type, \
                                  sizeof(type), rows, columns, description, status)
void * _mdarrays_load_2d_dynamic(char *cfile, uint cline, char *filename, 
                                 char *type_string, uint type_size, 
                                 uint *rows, uint *columns, char **description, uint *status);

#define MDARRAYS_PRINT_STYLE 0
#define MDARRAYS_SAVE_STYLE 1

// =================================
// Pretty Printing and styles
// =================================
#define MDARRAYS_SET_SAVE_STYLE_DEFAULTS(array) \
        _mdarrays_set_style_guts(__FILE__, __LINE__, MDARRAYS_SAVE_STYLE, array, true/*defaults*/, 0, 0, 0, 0, 0)
#define MDARRAYS_SET_SAVE_STYLE(array, delimiter, line_delimiter, type, width, precision) \
        _mdarrays_set_style_guts(__FILE__, __LINE__, MDARRAYS_SAVE_STYLE, array, false/*defaults*/, delimiter, \
                                 line_delimiter, type, width, precision)
#define MDARRAYS_SET_PRINT_STYLE_DEFAULTS(array) \
        _mdarrays_set_style_guts(__FILE__, __LINE__, MDARRAYS_PRINT_STYLE, array, true/*defaults*/, 0, 0, 0, 0, 0)
#define MDARRAYS_SET_PRINT_STYLE(array, delimiter, line_delimiter, type, width, precision) \
        _mdarrays_set_style_guts(__FILE__, __LINE__, MDARRAYS_PRINT_STYLE, array, \
                                 false/*defaults*/, delimiter, \
                                 line_delimiter, type, width, precision)
void _mdarrays_set_style_guts(char *cfile, uint cline, 
                              uint mode, void *array, bool defaults, char delimiter,
                              char line_delimiter, uint type, uint width, uint precision);

// =================================
// Abort function stuff
// =================================
#define _mdarrays_abort(file, line, ...)  \
    _mdarrays_abort_function(__FILE__, __LINE__, __func__, file, line, __VA_ARGS__)
void _mdarrays_abort_function(char *file_i, uint line_i, const char *function_i,
                              char *cfile, uint cline, char *fmt, ...);

// =================================
// Array Comparison
// =================================
#define MDARRAYS_EQUAL(array1, array2) \
    _mdarrays_equal(__FILE__, __LINE__, array1, array2)
bool _mdarrays_equal(char *cfile, uint cline, void *array1, void *array2);

// =================================
// Array Copy
// =================================
#define MDARRAYS_COPY_1D(from, from_offset, to, to_offset) \
    _mdarrays_copy_1d(__FILE__, __LINE__, from, from_offset, to, to_offset)
void _mdarrays_copy_1d(char *cfile, uint cline, void *from, uint from_offset,
                       void *to, uint to_offset);

#endif /* #ifndef _MDARRAYS_H */

